"""Dashboard for managed jobs based on Flask.

TODO(zongheng): This is a basic version. In the future we can beef up the web
frameworks used (e.g.,
https://github.com/ray-project/ray/tree/master/dashboard/client/src) and/or get
rid of the SSH port-forwarding business (see cli.py's job_dashboard()
comment).
"""
import collections
import datetime
import enum
import os
import pathlib

import flask
import yaml

from sky import jobs as managed_jobs
from sky.client import sdk
from sky.jobs import constants as managed_job_constants
from sky.utils import common_utils
from sky.utils import controller_utils

app = flask.Flask(__name__)


def _is_running_on_jobs_controller() -> bool:
    """Am I running on jobs controller?

    Loads ~/.sky/sky_ray.yml and check cluster_name.
    """
    if pathlib.Path('~/.sky/sky_ray.yml').expanduser().exists():
        config = yaml.safe_load(
            pathlib.Path('~/.sky/sky_ray.yml').expanduser().read_text(
                encoding='utf-8'))
        cluster_name = config.get('cluster_name', '')
        # We use startswith instead of exact match because the cluster name in
        # the yaml file is cluster_name_on_cloud which may have additional
        # suffices.
        return cluster_name.startswith(
            controller_utils.Controllers.JOBS_CONTROLLER.value.cluster_name)
    return False


# Column indices for job table
class JobTableColumns(enum.IntEnum):
    """Column indices for the jobs table in the dashboard.

    - DROPDOWN (0): Column for expandable dropdown arrow
    - ID (1): Job ID column
    - TASK (2): Task name/number column
    - NAME (3): Job name column
    - RESOURCES (4): Resources used by job
    - SUBMITTED (5): Job submission timestamp
    - TOTAL_DURATION (6): Total time since job submission
    - JOB_DURATION (7): Actual job runtime
    - RECOVERIES (8): Number of job recoveries
    - STATUS (9): Current job status
    - STARTED (10): Job start timestamp
    - CLUSTER (11): Cluster name
    - REGION (12): Cloud region
    - FAILOVER (13): Job failover history
    - DETAILS (14): Job details
    - ACTIONS (15): Available actions column
    - LOG_CONTENT (16): Log content column
    """
    DROPDOWN = 0
    ID = 1
    TASK = 2
    NAME = 3
    RESOURCES = 4
    SUBMITTED = 5
    TOTAL_DURATION = 6
    JOB_DURATION = 7
    RECOVERIES = 8
    STATUS = 9
    STARTED = 10
    CLUSTER = 11
    REGION = 12
    DETAILS = 13
    FAILOVER = 14
    ACTIONS = 15
    LOG_CONTENT = 16


# Column headers matching the indices above
JOB_TABLE_COLUMNS = [
    '', 'ID', 'Task', 'Name', 'Resources', 'Submitted', 'Total Duration',
    'Job Duration', 'Status', 'Started', 'Cluster', 'Region', 'Failover',
    'Recoveries', 'Details', 'Actions', 'Log Content'
]

# This column is given by format_job_table but should be ignored.
SCHED_STATE_COLUMN = 12


def _extract_launch_history(log_content: str) -> str:
    """Extract launch history from log content.

    Args:
        log_content: Content of the log file.
    Returns:
        A formatted string containing the launch history.
    """
    launches = []
    current_launch = None

    for line in log_content.splitlines():
        if 'Launching on' in line:
            try:
                parts = line.split(']')
                if len(parts) >= 2:
                    timestamp = parts[0].split()[1:3]
                    message = parts[1].replace('[0m⚙︎', '').strip()
                    formatted_line = f'{" ".join(timestamp)} {message}'
                    if current_launch:
                        prev_time, prev_target = current_launch.rsplit(
                            ' Launching on ', 1)
                        launches.append(
                            f'{prev_time} Tried to launch on {prev_target}')

                    # Store the current launch
                    current_launch = formatted_line
            except IndexError:
                launches.append(line.strip())

    # Add the final (successful) launch at the beginning
    if current_launch:
        result = [current_launch]
        result.extend(launches)
        return '\n'.join(result)

    return 'No launch history found'


@app.route('/')
def home():
    if not _is_running_on_jobs_controller():
        # Experimental: run on laptop (refresh is very slow).
        request_id = managed_jobs.queue(refresh=True, skip_finished=False)
        all_managed_jobs = sdk.get(request_id)
    else:
        job_table = managed_jobs.dump_managed_job_queue()
        all_managed_jobs = managed_jobs.load_managed_job_queue(job_table)

    timestamp = datetime.datetime.now(datetime.timezone.utc)
    rows = managed_jobs.format_job_table(all_managed_jobs,
                                         show_all=True,
                                         show_user=False,
                                         return_rows=True)

    status_counts = collections.defaultdict(int)
    for task in all_managed_jobs:
        if not task['status'].is_terminal():
            status_counts[task['status'].value] += 1

    # Add an empty column for the dropdown button and actions column
    # Exclude SCHED. STATE and LOG_CONTENT columns
    rows = [
        [''] + row[:SCHED_STATE_COLUMN] + row[SCHED_STATE_COLUMN + 1:] +
        # Add empty cell for failover and actions column
        [''] + [''] + [''] for row in rows
    ]

    # Add log content as a regular column for each job
    for row in rows:
        job_id = str(row[JobTableColumns.ID]).strip().replace(' ⤳', '')
        if job_id and job_id != '-':
            try:
                log_path = os.path.join(
                    os.path.expanduser(
                        managed_job_constants.JOBS_CONTROLLER_LOGS_DIR),
                    f'{job_id}.log')
                if os.path.exists(log_path):
                    with open(log_path, 'r', encoding='utf-8') as f:
                        log_content = f.read()
                        row[JobTableColumns.FAILOVER] = _extract_launch_history(
                            log_content)
                        row[JobTableColumns.LOG_CONTENT] = log_content
                else:
                    row[JobTableColumns.FAILOVER] = 'Log file not found'
                    row[JobTableColumns.LOG_CONTENT] = 'Log file not found'
            except (IOError, OSError) as e:
                error_msg = f'Error reading log: {str(e)}'
                row[JobTableColumns.FAILOVER] = error_msg
                row[JobTableColumns.LOG_CONTENT] = error_msg
        else:
            row[JobTableColumns.LOG_CONTENT] = ''

    if rows and len(rows[0]) != len(JOB_TABLE_COLUMNS):
        raise RuntimeError(
            f'Dashboard code and managed job queue code are out of sync. '
            f'Expected {JOB_TABLE_COLUMNS} columns, got {rows[0]}')

    # Fix STATUS color codes: '\x1b[33mCANCELLED\x1b[0m' -> 'CANCELLED'
    for row in rows:
        row[JobTableColumns.STATUS] = common_utils.remove_color(
            row[JobTableColumns.STATUS])

    # Remove filler rows ([''], ..., ['-'])
    rows = [
        row for row in rows
        if ''.join(map(str, row[:JobTableColumns.ACTIONS])) != ''
    ]

    # Get all unique status values
    status_values = sorted(
        list(set(row[JobTableColumns.STATUS] for row in rows)))

    rendered_html = flask.render_template(
        'index.html',
        columns=JOB_TABLE_COLUMNS,
        rows=rows,
        last_updated_timestamp=timestamp,
        status_values=status_values,
        status_counts=status_counts,
        request=flask.request,
    )
    return rendered_html


if __name__ == '__main__':
    app.run()



"""Utility functions for deploying Kubernetes clusters."""
import os
import shlex
import subprocess
import tempfile
from typing import List, Optional

from sky import check as sky_check
from sky import sky_logging
from sky.backends import backend_utils
from sky.clouds import cloud as sky_cloud
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.skylet import constants
from sky.skylet import log_lib
from sky.utils import log_utils
from sky.utils import rich_utils
from sky.utils import subprocess_utils
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)


def deploy_remote_cluster(ip_list: List[str],
                          ssh_user: str,
                          ssh_key: str,
                          cleanup: bool,
                          context_name: Optional[str] = None,
                          password: Optional[str] = None):
    success = False
    path_to_package = os.path.dirname(__file__)
    up_script_path = os.path.join(path_to_package, 'deploy_remote_cluster.sh')
    # Get directory of script and run it from there
    cwd = os.path.dirname(os.path.abspath(up_script_path))

    # Create temporary files for the IPs and SSH key
    with tempfile.NamedTemporaryFile(mode='w') as ip_file, \
         tempfile.NamedTemporaryFile(mode='w') as key_file:

        # Write IPs and SSH key to temporary files
        ip_file.write('\n'.join(ip_list))
        ip_file.flush()

        key_file.write(ssh_key)
        key_file.flush()
        os.chmod(key_file.name, 0o600)

        deploy_command = (f'{up_script_path} {ip_file.name} '
                          f'{ssh_user} {key_file.name}')
        if context_name is not None:
            deploy_command += f' {context_name}'
        if password is not None:
            deploy_command += f' --password {password}'
        if cleanup:
            deploy_command += ' --cleanup'

        # Convert the command to a format suitable for subprocess
        deploy_command = shlex.split(deploy_command)

        # Setup logging paths
        run_timestamp = sky_logging.get_run_timestamp()
        log_path = os.path.join(constants.SKY_LOGS_DIRECTORY, run_timestamp,
                                'local_up.log')

        if cleanup:
            msg_str = 'Cleaning up remote cluster...'
        else:
            msg_str = 'Deploying remote cluster...'
        with rich_utils.safe_status(
                ux_utils.spinner_message(msg_str,
                                         log_path=log_path,
                                         is_local=True)):
            returncode, _, stderr = log_lib.run_with_log(
                cmd=deploy_command,
                log_path=log_path,
                require_outputs=True,
                stream_logs=False,
                line_processor=log_utils.SkyRemoteUpLineProcessor(
                    log_path=log_path, is_local=True),
                cwd=cwd)
        if returncode == 0:
            success = True
        else:
            with ux_utils.print_exception_no_traceback():
                log_hint = ux_utils.log_path_hint(log_path, is_local=True)
                raise RuntimeError('Failed to deploy remote cluster. '
                                   f'Full log: {log_hint}'
                                   f'\nError: {stderr}')

        if success:
            if cleanup:
                logger.info(
                    ux_utils.finishing_message(
                        '🎉 Remote cluster cleaned up successfully.',
                        log_path=log_path,
                        is_local=True))
            else:
                logger.info(
                    ux_utils.finishing_message(
                        '🎉 Remote cluster deployed successfully.',
                        log_path=log_path,
                        is_local=True))


def deploy_local_cluster(gpus: bool):
    cluster_created = False

    # Check if GPUs are available on the host
    local_gpus_available = backend_utils.check_local_gpus()
    gpus = gpus and local_gpus_available

    # Check if ~/.kube/config exists:
    if os.path.exists(os.path.expanduser('~/.kube/config')):
        curr_context = kubernetes_utils.get_current_kube_config_context_name()
        skypilot_context = 'kind-skypilot'
        if curr_context is not None and curr_context != skypilot_context:
            logger.info(
                f'Current context in kube config: {curr_context}'
                '\nWill automatically switch to kind-skypilot after the local '
                'cluster is created.')
    message_str = 'Creating local cluster{}...'
    message_str = message_str.format((' with GPU support (this may take up '
                                      'to 15 minutes)') if gpus else '')
    path_to_package = os.path.dirname(__file__)
    up_script_path = os.path.join(path_to_package, 'create_cluster.sh')

    # Get directory of script and run it from there
    cwd = os.path.dirname(os.path.abspath(up_script_path))
    run_command = up_script_path + ' --gpus' if gpus else up_script_path
    run_command = shlex.split(run_command)

    # Setup logging paths
    run_timestamp = sky_logging.get_run_timestamp()
    log_path = os.path.join(constants.SKY_LOGS_DIRECTORY, run_timestamp,
                            'local_up.log')
    logger.info(message_str)

    with rich_utils.safe_status(
            ux_utils.spinner_message(message_str,
                                     log_path=log_path,
                                     is_local=True)):
        returncode, _, stderr = log_lib.run_with_log(
            cmd=run_command,
            log_path=log_path,
            require_outputs=True,
            stream_logs=False,
            line_processor=log_utils.SkyLocalUpLineProcessor(log_path=log_path,
                                                             is_local=True),
            cwd=cwd)

    # Kind always writes to stderr even if it succeeds.
    # If the failure happens after the cluster is created, we need
    # to strip all stderr of "No kind clusters found.", which is
    # printed when querying with kind get clusters.
    stderr = stderr.replace('No kind clusters found.\n', '')

    if returncode == 0:
        cluster_created = True
    elif returncode == 100:
        logger.info(
            ux_utils.finishing_message(
                'Local cluster already exists.\n',
                log_path=log_path,
                is_local=True,
                follow_up_message=
                'If you want to delete it instead, run: sky local down'))
    else:
        with ux_utils.print_exception_no_traceback():
            log_hint = ux_utils.log_path_hint(log_path, is_local=True)
            raise RuntimeError('Failed to create local cluster. '
                               f'Full log: {log_hint}'
                               f'\nError: {stderr}')
    # Run sky check
    with rich_utils.safe_status('[bold cyan]Running sky check...'):
        sky_check.check_capability(sky_cloud.CloudCapability.COMPUTE,
                                   quiet=True,
                                   clouds=['kubernetes'])
    if cluster_created:
        # Prepare completion message which shows CPU and GPU count
        # Get number of CPUs
        p = subprocess_utils.run(
            'kubectl get nodes -o jsonpath=\'{.items[0].status.capacity.cpu}\'',
            capture_output=True)
        num_cpus = int(p.stdout.decode('utf-8'))

        # GPU count/type parsing
        gpu_message = ''
        gpu_hint = ''
        if gpus:
            # Get GPU model by querying the node labels
            label_name_escaped = 'skypilot.co/accelerator'.replace('.', '\\.')
            gpu_type_cmd = f'kubectl get node skypilot-control-plane -o jsonpath=\"{{.metadata.labels[\'{label_name_escaped}\']}}\"'  # pylint: disable=line-too-long
            try:
                # Run the command and capture the output
                gpu_count_output = subprocess.check_output(gpu_type_cmd,
                                                           shell=True,
                                                           text=True)
                gpu_type_str = gpu_count_output.strip() + ' '
            except subprocess.CalledProcessError as e:
                output = str(e.output.decode('utf-8'))
                logger.warning(f'Failed to get GPU type: {output}')
                gpu_type_str = ''

            # Get number of GPUs (sum of nvidia.com/gpu resources)
            gpu_count_command = 'kubectl get nodes -o=jsonpath=\'{range .items[*]}{.status.allocatable.nvidia\\.com/gpu}{\"\\n\"}{end}\' | awk \'{sum += $1} END {print sum}\''  # pylint: disable=line-too-long
            try:
                # Run the command and capture the output
                gpu_count_output = subprocess.check_output(gpu_count_command,
                                                           shell=True,
                                                           text=True)
                gpu_count = gpu_count_output.strip(
                )  # Remove any extra whitespace
                gpu_message = f' and {gpu_count} {gpu_type_str}GPUs'
            except subprocess.CalledProcessError as e:
                output = str(e.output.decode('utf-8'))
                logger.warning(f'Failed to get GPU count: {output}')
                gpu_message = f' with {gpu_type_str}GPU support'

            gpu_hint = (
                '\nHint: To see the list of GPUs in the cluster, '
                'run \'sky show-gpus --cloud kubernetes\'') if gpus else ''

        if num_cpus < 2:
            logger.info('Warning: Local cluster has less than 2 CPUs. '
                        'This may cause issues with running tasks.')
        logger.info(
            ux_utils.finishing_message(
                message=(f'Local Kubernetes cluster created successfully with '
                         f'{num_cpus} CPUs{gpu_message}.'),
                log_path=log_path,
                is_local=True,
                follow_up_message=(
                    '\n`sky launch` can now run tasks locally.\n'
                    'Hint: To change the number of CPUs, change your docker '
                    'runtime settings. See https://kind.sigs.k8s.io/docs/user/quick-start/#settings-for-docker-desktop for more info.'  # pylint: disable=line-too-long
                    f'{gpu_hint}')))
